import java.io.IOException;
import java.io.InputStream;

public class Convertor {
	
	private byte[] big5bytes;
	private byte[] ucs2bytes;
	private char[] ucs2chars;

	public void ucs2ToBig5Bytes( char c, byte[] b )
	{
		// 理論上若 c < 0x80 是不應該進來這個函數的 
		if( c < 0x80 ) {
			b[0] = (byte) c;
			return;
		}
		
		// 表是從 Unicode 0x80 開始建的
		b[0] = big5bytes[ (c - 0x80) * 2 ];
		b[1] = big5bytes[ (c - 0x80) * 2 + 1 ];
	}

	public char big5BytesToUCS2( byte b1, byte b2 )
	{
		// signed to unsigned
		int i1 = (b1 < 0 ? 256 : 0) + (int)b1;
		int i2 = (b2 < 0 ? 256 : 0) + (int)b2; 

		// 表是從 big5 0x8140 開始建的
		int shift = ((i1 << 8) | i2) - 0x8140;
		
		// 超過 big5 的範圍
		if( shift < 0 || shift * 2 + 1 >= ucs2bytes.length ) {
			return '?';
		}
		
		i1 = (ucs2bytes[shift * 2] < 0 ? 256 : 0) + (int)ucs2bytes[shift * 2];
		i2 = (ucs2bytes[shift * 2 + 1] < 0 ? 256 : 0) + (int)ucs2bytes[shift * 2 + 1];
		
		return (char) (i1 << 8 | i2);
	}
	
	// 把檔案讀進 byte array
	private int readFile( String name, byte[] b )
	{
		int begin = 0, len;
		InputStream is = ZTerm.class.getResourceAsStream( name );
		
		try {
			// XXX: 本來應該 is.read( b ) 就可以才對，
			// 但是我發現在包裝成 jar 以後就會讀不完整，一定要這樣讀。
			while( true ) {
				len = is.read( b, begin, b.length - begin );
				if( len == -1 ) {
					break;
				}
				begin += len;
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		return begin;
	}
	
	public Convertor()
	{
		int i1, i2;
		
		ucs2bytes = new byte[64 * 1024];
		big5bytes = new byte[128 * 1024];
		
		readFile( "conv/ucs2.txt", ucs2bytes );
		readFile( "conv/big5.txt", big5bytes );
		
		ucs2chars = new char[ ucs2bytes.length / 2 ];
		
		// 把讀進來的 ucs2 bytes 處理成標準的 (ucs2) char
		for( int i = 0; i < ucs2bytes.length; i += 2) {
			i1 = (ucs2bytes[i] < 0 ? 256 : 0) + (int)ucs2bytes[i];
			i2 = (ucs2bytes[i + 1] < 0 ? 256 : 0) + (int)ucs2bytes[i + 1];
			ucs2chars[i / 2] = (char)( i1 << 8 | i2 );
		}		
	}
}
