import java.awt.BorderLayout;
import java.awt.DefaultKeyboardFocusManager;
import java.awt.FlowLayout;
import java.awt.KeyEventDispatcher;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTabbedPane;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

public class ZTerm extends JFrame implements ActionListener, ChangeListener, KeyEventDispatcher, KeyListener, ComponentListener {
	
	private int locationx, locationy;
	private int width, height;
	
	private JTabbedPane tabbedPane;
	private JPanel controlPanel;
		
	private JMenuBar menuBar;
	private JMenu connectMenu, favoriteMenu, editMenu, optionMenu, helpMenu;
	private JMenuItem openItem, closeItem, reopenItem, quitItem;
	private JMenuItem copyItem, pasteItem, colorCopyItem, colorPasteItem;
	private JMenuItem preferenceItem, siteManagerItem, showToolbarItem;
	private JMenuItem usageItem, faqItem, aboutItem;
	private JMenuItem[] favoriteItems;
	
	private JPopupMenu popupMenu;
	private JMenuItem popupCopyItem, popupPasteItem, popupColorCopyItem, popupColorPasteItem, popupCopyLinkItem;
	
	private JToolBar connectionToolBar;
	private JButton openButton, closeButton, reopenButton;
	private JButton copyButton, colorCopyButton, pasteButton, colorPasteButton;
	private JButton telnetButton, sshButton;
	private JLabel siteLabel;
	private JTextField siteField;
	
	private ImageIcon connectedIcon, closedIcon, bellIcon;
	
	private Resource resource;
	private Vector sessions;
	private Clip clip;
	private String colorText = null;
	
	// 是否顯示工具列
	private boolean showToolbar;
	
	// 按滑鼠右鍵時滑鼠下的連結
	private String tmpLink;
	
	private BufferedImage bi;
	private Convertor conv;
	
	private boolean isShowingDialog;
	
	public static final int ICON_CONNECTED = 1;
	public static final int ICON_CLOSED = 2;
	public static final int ICON_ALERT = 3;
	
	// 建立各種選單，包括 popup 選單也在這裡建立
	private void makeMenu()
	{
		menuBar = new JMenuBar();
		
		connectMenu = new JMenu("連線");
		connectMenu.setMnemonic(KeyEvent.VK_N);
		connectMenu.setToolTipText("meta-n");

		favoriteMenu = new JMenu("我的最愛");
		favoriteMenu.setMnemonic(KeyEvent.VK_F);
		favoriteMenu.setToolTipText("meta-f");
		
		editMenu = new JMenu("編輯");
		editMenu.setMnemonic(KeyEvent.VK_E);
		editMenu.setToolTipText("meta-e");
		
		optionMenu = new JMenu("選項");
		optionMenu.setMnemonic(KeyEvent.VK_O);
		optionMenu.setToolTipText("meta-o");
		
		helpMenu = new JMenu("求助");
		helpMenu.setMnemonic(KeyEvent.VK_H);		
		helpMenu.setToolTipText("meta-h");
		
		popupMenu = new JPopupMenu();
		
		openItem = new JMenuItem("開新連線");
		openItem.setToolTipText("meta-q");
		openItem.addActionListener( this );
		
		closeItem = new JMenuItem("中斷連線");
		closeItem.setToolTipText("meta-w");
		closeItem.addActionListener( this );

		reopenItem = new JMenuItem("重新連線");
		reopenItem.setToolTipText("meta-r");
		reopenItem.addActionListener( this );
		
		quitItem = new JMenuItem("離開");
		quitItem.addActionListener( this );

		copyItem = new JMenuItem("複製");
		copyItem.addActionListener( this );
		copyItem.setToolTipText("meta-c");
		
		pasteItem = new JMenuItem("貼上");
		pasteItem.addActionListener( this );
		pasteItem.setToolTipText("meta-v");

		colorCopyItem = new JMenuItem("彩色複製");
		colorCopyItem.addActionListener( this );
		
		colorPasteItem = new JMenuItem("彩色貼上");
		colorPasteItem.addActionListener( this );
		
		preferenceItem = new JMenuItem("偏好設定");
		preferenceItem.addActionListener( this );
		preferenceItem.setToolTipText("meta-,");
		
		siteManagerItem = new JMenuItem("站台管理");
		siteManagerItem.addActionListener( this );
		siteManagerItem.setToolTipText("meta-.");
		
		showToolbarItem = new JMenuItem( showToolbar ? "隱藏工具列" : "顯示工具列" );
		showToolbarItem.addActionListener( this );
		
		usageItem = new JMenuItem("使用說明");
		usageItem.addActionListener( this );
		
		faqItem = new JMenuItem( "常見問題" );
		faqItem.addActionListener( this );

		aboutItem = new JMenuItem("關於 ZTerm");
		aboutItem.addActionListener( this );
		
		setJMenuBar( menuBar);
		
		menuBar.add( connectMenu );
		menuBar.add( favoriteMenu );
		menuBar.add( editMenu );
		menuBar.add( optionMenu );
		menuBar.add( helpMenu );
		
		connectMenu.add( openItem );
		connectMenu.add( closeItem );
		connectMenu.add( reopenItem );
		connectMenu.add( quitItem );
		
		updateFavoriteMenu();
		
		editMenu.add( copyItem );
		editMenu.add( pasteItem );
		editMenu.add( colorCopyItem );
		editMenu.add( colorPasteItem );
		
		optionMenu.add( preferenceItem );
		optionMenu.add( siteManagerItem );
		optionMenu.add( showToolbarItem );
		
		helpMenu.add( usageItem );
		helpMenu.add( faqItem );
		helpMenu.add( aboutItem );
	}

	private void makePopupMenu()
	{
		popupCopyItem = new JMenuItem("複製");
		popupCopyItem.addActionListener( this );

		popupColorCopyItem = new JMenuItem("彩色複製");
		popupColorCopyItem.addActionListener( this );
		
		popupPasteItem = new JMenuItem("貼上");
		popupPasteItem.addActionListener( this );
		
		popupColorPasteItem = new JMenuItem("彩色貼上");
		popupColorPasteItem.addActionListener( this );
		
		popupCopyLinkItem = new JMenuItem("複製連結");
		popupCopyLinkItem.addActionListener( this );
		popupCopyLinkItem.setEnabled( false );
		
		popupMenu.add( popupCopyItem );
		popupMenu.add( popupPasteItem );
		popupMenu.add( popupColorCopyItem );
		popupMenu.add( popupColorPasteItem );
		popupMenu.add( popupCopyLinkItem );
	}
	
	private void makeTabbedPane()
	{
		// tab 擺在上面，太多 tab 時使用捲頁的顯示方式
		tabbedPane = new JTabbedPane( JTabbedPane.TOP, JTabbedPane.SCROLL_TAB_LAYOUT );
		tabbedPane.addChangeListener( this );
		getContentPane().add( tabbedPane );
	}
	
	private void makeControlPanel()
	{
		controlPanel = new JPanel();
		controlPanel.setLayout( new FlowLayout( FlowLayout.LEFT ) );
		
		connectionToolBar = new JToolBar();
		connectionToolBar.setVisible( showToolbar );
		
		closeButton = new JButton( "斷線");
		closeButton.setToolTipText("meta-w");
		closeButton.setFocusable( false );
		closeButton.addActionListener( this );
		
		reopenButton = new JButton( "重連" );
		reopenButton.setToolTipText("meta-r");
		reopenButton.setFocusable( false );
		reopenButton.addActionListener( this );
		
		copyButton = new JButton("複製");
		copyButton.setToolTipText("meta-c");
		copyButton.setFocusable( false );
		copyButton.addActionListener( this );
		
		pasteButton = new JButton("貼上");
		pasteButton.setToolTipText("meta-v");
		pasteButton.setFocusable( false );
		pasteButton.addActionListener( this );
		
		colorCopyButton = new JButton("彩色複製");
		colorCopyButton.setFocusable( false );
		colorCopyButton.addActionListener( this );
		
		colorPasteButton = new JButton("彩色貼上");
		colorPasteButton.setFocusable( false );
		colorPasteButton.addActionListener( this );
		
		telnetButton = new JButton("Telnet");
		telnetButton.setToolTipText("meta-t");
		telnetButton.setFocusable( false );
		telnetButton.addActionListener( this );
		
		sshButton = new JButton("SSH");
		sshButton.setToolTipText("meta-s");
		sshButton.setFocusable( false );
		sshButton.addActionListener( this );
		
		siteLabel = new JLabel("位置");
		siteLabel.setFocusable( false );
		
		siteField = new JTextField( 32 );
		siteField.setToolTipText("meta-d");
		siteField.addActionListener( this );
		siteField.addKeyListener( this );
		
		openButton = new JButton( "連線" );
		openButton.setFocusable( false );
		openButton.addActionListener( this );
		
		connectionToolBar.add(closeButton);
		connectionToolBar.add(reopenButton);
		
		connectionToolBar.add( new JToolBar.Separator() );
		
		connectionToolBar.add(copyButton);
		connectionToolBar.add(pasteButton);
		connectionToolBar.add(colorCopyButton);
		connectionToolBar.add(colorPasteButton);
		
		connectionToolBar.add( new JToolBar.Separator() );
		
		connectionToolBar.add(telnetButton);
		connectionToolBar.add(sshButton);
		
		connectionToolBar.add( new JToolBar.Separator() );
		
		connectionToolBar.add( siteLabel );
		connectionToolBar.add( siteField );
		connectionToolBar.add( openButton );
		
		controlPanel.add( connectionToolBar );

		getContentPane().add( controlPanel, BorderLayout.NORTH );
	}
	
	public void updateFavoriteMenu()
	{
		Site fa;
		Vector f = resource.getFavorites();
		favoriteItems = new JMenuItem[ f.size() ];

		favoriteMenu.removeAll();

		// 顯示目前我的最愛內容
		for( int i = 0; i < f.size(); i++ ) {
			fa = (Site)f.elementAt(i);
			favoriteItems[i] = new JMenuItem( fa.name );
			favoriteItems[i].setToolTipText( fa.host + ":" + fa.port );
			favoriteItems[i].addActionListener( this );
			favoriteMenu.add( favoriteItems[i] );
		}
	}
	
	public void updateToolbar()
	{
		showToolbar = resource.getBooleanValue( Resource.SHOW_TOOLBAR );
		showToolbarItem.setText( showToolbar ? "隱藏工具列" : "顯示工具列" );
		connectionToolBar.setVisible( showToolbar );
	}
	
	public void updateBounds()
	{
		locationx = resource.getIntValue( Resource.GEOMETRY_X );
		locationy = resource.getIntValue( Resource.GEOMETRY_Y );
		width = resource.getIntValue( Resource.GEOMETRY_WIDTH );
		height = resource.getIntValue( Resource.GEOMETRY_HEIGHT );

		setBounds( locationx, locationy, width, height );
		validate();
	}
	
	public void updateSize()
	{
		Session s;
		
		width = getWidth();
		height = getHeight();		
		
		// 產生跟主視窗一樣大的 image
		bi = new BufferedImage( width, height, BufferedImage.TYPE_INT_RGB );
		
		// 視窗大小調整時同步更新每個 session 的大小
		for( int i = 0; i < sessions.size(); i++ ) {
			s = (Session)sessions.elementAt( i ); 
			s.updateImage( bi );
			s.updateSize();
			s.updateFont();
			s.updateScreen();
		}
	}

	public void updateAntiIdleTime()
	{
		for( int i = 0; i < sessions.size(); i++ ) {
			((Session)sessions.elementAt( i )).updateAntiIdleTime();
		}
	}
	
	public void updateLookAndFeel()
	{
		try {
			// 使用系統的 Look and Feel
			if( resource.getBooleanValue( Resource.SYSTEM_LOOK_FEEL ) ) {
				UIManager.setLookAndFeel( UIManager.getSystemLookAndFeelClassName() );
			} else {
				UIManager.setLookAndFeel( UIManager.getCrossPlatformLookAndFeelClassName() );
			}
			SwingUtilities.updateComponentTreeUI( this );
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (UnsupportedLookAndFeelException e) {
			e.printStackTrace();
		}
	}
	
	public void updateTab()
	{
		// 為了下面 invokeLater 的關係，這邊改成 final
		final Session s = (Session)tabbedPane.getSelectedComponent();
		
		if( s != null ) {
			
			// 修改視窗標題列
			setTitle( "ZTerm - " + s.getWindowTitle() );
			
			// 修改網址列
			siteField.setText( s.getURL() );
			
			// 切換到已連線的 session 時設定 icon 為 connected, 以取消  bell icon.
			if( s.isClosed() ) {
				setIcon( ICON_CLOSED, s );
			} else {
				setIcon( ICON_CONNECTED, s );
			}
			
			// 因為全部的連線共用一張 BufferedImage, 切換分頁時需重繪內容。
			s.updateScreen();
			
			// 讓所選的 session 取得 focus
			// XXX: s.requestFocusInWindow(); 無法正確取得 focus（似乎是 Java 的 bug）
			// 因此必須使用下面的 workaround
	        javax.swing.SwingUtilities.invokeLater(new Runnable() {
	            public void run() {
	            	s.requestFocusInWindow();
	            }
	        });
	        
		} else {
			// FIXME: magic number
			setTitle("ZTerm");
			siteField.setText( "" );
		}
	}
	
	public void updateTabTitle()
	{
		for( int i = 0; i < tabbedPane.getTabCount(); i++) {
			tabbedPane.setTitleAt( i, (i + 1) + ". " + ((Session)sessions.elementAt(i)).getHost() );
		}
	}
	
	public void changeSession( int index )
	{
		if( 0 <= index && index < tabbedPane.getTabCount() ) {
			tabbedPane.setSelectedIndex( index );
			// System.out.println("Change to session: " + index );
		} else {
			// System.out.println("Change to session: " + index + ", error range!");
		}
	}

	public void showMessage(String msg)
	{
		JOptionPane.showMessageDialog( this, msg );
	}
	
	public void showPopup(int x, int y, String link )
	{

		Point p1 = tabbedPane.getLocation();
		Point p2 = tabbedPane.getSelectedComponent().getLocation();
		
		// TODO: 用 tmpLink 的作法蠻笨的，但暫時想不到好作法
		popupCopyLinkItem.setEnabled( link != null );
		tmpLink = link;
		
		popupMenu.show( this, p1.x + p2.x + x, p1.y + p2.y + y );
	}
	
	private void autoconnect()
	{
		Vector f = resource.getFavorites();
		Iterator iter = f.iterator();
		Site s;
		
		while( iter.hasNext() ) {
			s = (Site)iter.next();
			if( s.autoconnect ) {
				connect( s.host, s.port, s.protocol, -1 );
			}
		}
	}
	
	private void connect( String h )
	{
		Site s;
		String host;
   		int port, pos, prot;

   		// 如果開新連線時按了取消則傳回值為 null
		if( h == null || h.length() == 0 ) {
			return;
		}
        
		do {
			// 透過 name or alias 連線
			s = resource.getFavorite( h );
			if( s != null ) {
				host = s.host;
				port = s.port;
				prot = s.protocol;
				break;
			}

			pos = h.indexOf( "://" );
			// Default 就是 telnet
			prot = Network.TELNET;
			if( pos != -1 ) {
				if( h.substring( 0, pos).equalsIgnoreCase("ssh") ) {
					prot = Network.SSH;
				} else if( h.substring( 0, pos).equalsIgnoreCase("telnet") ) {
					prot = Network.TELNET;
				} else {
					showMessage("通訊協定錯誤！");
					return;
				}
				// 將 h 重設為 :// 後的東西
				h = h.substring( pos + 3 );
			}
			
			// 取得 host:port, 或 host(:23)
			pos = h.indexOf(':');
			if( pos == -1 ) {
				host = h;
				if( prot == Network.TELNET ) {
					port = 23;
				} else {
					port = 22;
				}
			} else {
				host = h.substring( 0, pos );
				port = Integer.parseInt( h.substring( pos + 1) );
			}
		} while( false );
		
		// host 長度為零則不做事
		if( h.length() == 0 ) {
			return;
		}
		
		connect( host, port, prot, -1 );
	}

	private void connect( String host, int port, int protocol, int index )
	{
		Session s;
		
		// 大小由 tab 的可視範圍決定。
		// 但實際上每個 tab 可視範圍也沒這麼大，Session 內部會再自動調整（updateSize）
		Rectangle r = tabbedPane.getVisibleRect();
		s = new Session( host, port, protocol, this, r.width, r.height, bi, resource, conv );
		
		// index 為連線後放在第幾個分頁，若為 -1 表開新分頁。
		if( index == -1 ) {
			sessions.add( s );
			tabbedPane.addTab( (tabbedPane.getTabCount() + 1) + ". " + host, closedIcon, s, host );
			tabbedPane.setSelectedIndex( tabbedPane.getTabCount() - 1);
		} else {
			sessions.setElementAt( s, index );
			tabbedPane.setComponentAt( index, s);
		}

		// 每個 session 都是一個 thread, 解決主程式被 block 住的問題。
		// 在 Session 的 run() 中會取得 Focus
		new Thread( s ).start();
	}
	
	public void openExternalBrowser( String url ) {
		int pos;
		
		String cmd = resource.getStringValue( Resource.EXTERNAL_BROWSER );
		if( cmd == null ) {
			showMessage( "瀏覽器啟動命令錯誤！" );
			return;
		}
		
		// 把 %u 置換成給定的 url
		pos = cmd.indexOf( "%u" );
		if( pos == -1 ) {
			showMessage( "瀏覽器啟動命令錯誤！" );
			return;
		}
		cmd = cmd.substring( 0, pos ) + url + cmd.substring( pos + 2 );
		// System.out.println( "browser command: " + cmd );
		
		runExternal( cmd );
	}
	
	public void runExternal( String cmd ) {
		try {
			Runtime.getRuntime().exec( cmd );
		} catch (IOException e) {
			e.printStackTrace();
		}	
	}
	
	public void bell()
	{	
		if( resource.getBooleanValue( Resource.USE_CUSTOM_BELL) ) {
			try {
				java.applet.Applet.newAudioClip( new File( resource.getStringValue( Resource.CUSTOM_BELL_PATH ) ).toURL() ).play();
			} catch (MalformedURLException e) {
				e.printStackTrace();
			}
		} else {
			java.awt.Toolkit.getDefaultToolkit().beep();
		}
		
		String os = System.getProperty("os.name");
		if( os.startsWith("Linux") ) {
			requestFocus();
			updateTab();
		} else if( os.startsWith("Mac") ) {
			new MacAlerter().alert();
		} else if( os.startsWith("Windows") ) {
			new WindowsAlerter().alert( this );
		}
	}
	
	public void open()
	{
		String host = JOptionPane.showInputDialog( this, "請輸入連線位置：");
		connect(host);
	}
	
	public void quit()
	{
		for( int i = 0; i < sessions.size(); i++) {
			if( !((Session)sessions.elementAt(i)).isClosed() ) {
				isShowingDialog = true;
				if( JOptionPane.showConfirmDialog( this, "還有連線尚未中斷，您確定要關閉程式嗎？", "連線中", JOptionPane.YES_NO_OPTION ) != JOptionPane.YES_OPTION ) {
					isShowingDialog = false;
					return;
				} else {
					isShowingDialog = false;
					break;
				}
			}
		}

		for( int i = 0; i < sessions.size(); i++) {
			((Session)sessions.elementAt(i)).close( false );
		}
		sessions.removeAllElements();

		// 紀錄結束時的視窗大小
		Rectangle r = getBounds();
		resource.setValue( Resource.GEOMETRY_X, r.x );
		resource.setValue( Resource.GEOMETRY_Y, r.y );
		resource.setValue( Resource.GEOMETRY_WIDTH, r.width );
		resource.setValue( Resource.GEOMETRY_HEIGHT, r.height );
		
		// 程式結束，把目前設定值寫回設定檔。
		resource.writeFile();
		
		System.exit( 0 );
	}
	
	public void copy()
	{
		Session s = (Session)tabbedPane.getSelectedComponent();
		
		if( s != null ) {
			String str = s.getSelectedText();
			if( str.length() != 0 ) {
				clip.setContent( str );
			}
			if( resource.getBooleanValue( Resource.CLEAR_AFTER_COPY) ) {
				s.resetSelected();
				s.repaint();
			}
		}
	}
	
	public void colorCopy()
	{
		Session s = (Session)tabbedPane.getSelectedComponent();
		
		if( s != null ) {
			String str = s.getSelectedColorText();
			if( str.length() != 0 ) {
				colorText = str;
			}
			if( resource.getBooleanValue( Resource.CLEAR_AFTER_COPY) ) {
				s.resetSelected();
				s.repaint();
			}
		}
	}

	public void copyLink()
	{
		if( tmpLink != null ) {
			clip.setContent( tmpLink );
		}
	}
	
	public void paste()
	{
		Session s = (Session)tabbedPane.getSelectedComponent();
		if( s != null ) {
			s.pasteText( clip.getContent() );
		}
	}
	
	public void colorPaste()
	{
		Session s = (Session)tabbedPane.getSelectedComponent();
		if( s != null && colorText != null ) {
			s.pasteColorText( colorText );
		}
	}
	
	public void setIcon( int icon, Session s )
	{
		int index;
		ImageIcon ii;
		
		switch( icon ) {
			case ICON_CONNECTED:
				ii = connectedIcon;
				break;
			case ICON_CLOSED:
				ii = closedIcon;
				break;
			case ICON_ALERT:
				ii = bellIcon;
				break;
			default:
				ii = null;
		}
		
		index = tabbedPane.indexOfComponent( s );
		if( index != -1 ) {
			tabbedPane.setIconAt( index, ii );
		}
	}
	
	public boolean isTabForeground( Session s )
	{
		return (tabbedPane.indexOfComponent(s) == tabbedPane.getSelectedIndex());
	}
	
	public void closeCurrentTab()
	{
		Session s = (Session)tabbedPane.getSelectedComponent();

		if( s != null ) {
			
			// 連線中則詢問是否要斷線
			if( !s.isClosed() ) {
				isShowingDialog = true;
				if( JOptionPane.showConfirmDialog( this, "您確定要結束連線嗎？", "連線中", JOptionPane.YES_NO_OPTION ) != JOptionPane.YES_OPTION ) {
					isShowingDialog = false;
					return;
				}
				isShowingDialog = false;
			
				// 通知 session 要中斷連線了
				s.close( false );
				
				if( !resource.getBooleanValue( Resource.REMOVE_MANUAL_DISCONNECT ) ) {
					return;
				}
			}

			// 通知 session 要被移除了
			s.remove();
			
			tabbedPane.remove( s );
			sessions.remove( s );
			
			// 刪除分頁會影響分頁編號
			updateTabTitle();
			
			// 讓現在被選取的分頁取得 focus.
			updateTab();
		} else {
			// 沒有分頁了，關閉程式
			quit();
		}
	}
	
	public void reopenSession( Session s )
	{
		if( s != null ) {
			// 若連線中則開新分頁，已斷線則重連。
			if( s.isClosed() ) {
				connect( s.getHost(), s.getPort(), s.getProtocol(), tabbedPane.indexOfComponent(s) );
			} else {
				connect( s.getHost(), s.getPort(), s.getProtocol(), -1 );
			}
		}
	}
	
	private void showPreference()
	{
		new Preference( this, resource );
	}
	
	private void showSiteManager()
	{
		new SiteManager( this, resource );
	}
	
	private void showUsage()
	{
		new HtmlDialog( this, "使用說明", ZTerm.class.getResource( "docs/usage.html" ) );
	}
	
	private void showFAQ()
	{
		new HtmlDialog( this, "常見問題", ZTerm.class.getResource( "docs/faq.html" ) );
	}
	
	private void showAbout()
	{
		// FIXME: magic number
		JOptionPane.showMessageDialog( this, "ZTerm 0.32 版\n專案網頁：http://zhouer.org/ZTerm/" );
	}
	
	public void actionPerformed( ActionEvent ae ) {
		
		Object source = ae.getSource();
		
		if( source == openItem ) {
			open();
		} else if( source == siteField || source == openButton ) {
			String host = siteField.getText();
			connect( host );
		} else if( source == closeItem || source == closeButton ) {
			closeCurrentTab();
		} else if( source == reopenItem || source == reopenButton ) {
			reopenSession( (Session)tabbedPane.getSelectedComponent() );
		} else if( source == quitItem ) {
			quit();
		} else if( source == copyItem || source == copyButton || source == popupCopyItem ) {
			copy();
		} else if( source == colorCopyItem || source == colorCopyButton || source == popupColorCopyItem ) {
			colorCopy();
		} else if( source == pasteItem || source == pasteButton || source == popupPasteItem ) {
			paste();
		} else if( source == colorPasteItem || source == colorPasteButton || source == popupColorPasteItem ) {
			colorPaste();
		} else if( source == popupCopyLinkItem ) {
			copyLink();
		} else if( source == telnetButton ) {
			siteField.setText( "telnet://" );
			siteField.requestFocusInWindow();
		} else if( source == sshButton ) {
			siteField.setText( "ssh://" );
			siteField.requestFocusInWindow();
		} else if( source == preferenceItem ) {
			showPreference();
		} else if( source == siteManagerItem ) {
			showSiteManager();
		} else if( source == showToolbarItem ) {
			showToolbar = !showToolbar;
			resource.setValue( Resource.SHOW_TOOLBAR, showToolbar);
			updateToolbar();
			updateSize();
		} else if( source == usageItem ) {
			showUsage();
		} else if( source == faqItem ) {
			showFAQ();
		} else if( source == aboutItem ) {
			showAbout();
		}
		
		// 我的最愛列表
		for( int i = 0; i < favoriteItems.length; i++ ) {
			if( source == favoriteItems[i] ) {
				Site f = resource.getFavorite(favoriteItems[i].getText());
				connect( f.host, f.port, f.protocol, -1 );
			}
		}
	}
	
	public void stateChanged(ChangeEvent e) {
		// 切換分頁，更新視窗標題、畫面
		if( e.getSource() == tabbedPane ) {
			updateTab();
		}
	}
	
	public boolean dispatchKeyEvent( KeyEvent ke ) {

		// 只處理按下的狀況
		if( ke.getID() != KeyEvent.KEY_PRESSED ) {
			return false;
		}
		
		// 如果正在顯示 dialog 則不要處理
		// 原本是期望 modal dialog 會 block 住 keyboard event 的，不過看來還是得自己判斷
		if( isShowingDialog ) {
			return false;
		}
		
		if( ke.isAltDown() || ke.isMetaDown() ) {
		
			if( ke.getKeyCode() == KeyEvent.VK_C ) {
				copy();
			} else if( ke.getKeyCode() == KeyEvent.VK_D ) {
				siteField.requestFocusInWindow();
				siteField.selectAll();
			} else if( ke.getKeyCode() == KeyEvent.VK_Q ) {
				open();
			} else if( ke.getKeyCode() == KeyEvent.VK_R ) {
				reopenSession( (Session)tabbedPane.getSelectedComponent() );
			} else if( ke.getKeyCode() == KeyEvent.VK_S ) {
				siteField.setText("ssh://");
				siteField.requestFocusInWindow();
			} else if( ke.getKeyCode() == KeyEvent.VK_T ) {
				siteField.setText("telnet://");
				siteField.requestFocusInWindow();
			} else if( ke.getKeyCode() == KeyEvent.VK_V ) {
				paste();
			} else if( ke.getKeyCode() == KeyEvent.VK_W ) {
				closeCurrentTab();
			} else if( KeyEvent.VK_0 <= ke.getKeyCode() && ke.getKeyCode() <= KeyEvent.VK_9 ) {
				// 切換 tab 快速建
				// 可用 alt-N 或是 cmd-N 切換
				// XXX: 其實可以考慮 alt-0 不是快速建，不然多一個判斷很不舒服。
			
				if( KeyEvent.VK_0 == ke.getKeyCode() ) {
					changeSession( 9 );
				} else {
					changeSession( ke.getKeyCode() - KeyEvent.VK_1 );
				}
			} else if( ke.getKeyCode() == KeyEvent.VK_LEFT || ke.getKeyCode() == KeyEvent.VK_Z ) {
				// alt-left 切到上一個連線視窗
				// index 是否合法在 changeSession 內會判斷
				
				changeSession( tabbedPane.getSelectedIndex() - 1 );
			} else if( ke.getKeyCode() == KeyEvent.VK_RIGHT || ke.getKeyCode() == KeyEvent.VK_X ) {
				// alt-right 切到下一個連線視窗
				// index 是否合法在 changeSession 內會判斷
				
				changeSession( tabbedPane.getSelectedIndex() + 1 );
			} else if( ke.getKeyCode() == KeyEvent.VK_COMMA ) {
				// alt-, 開啟偏好設定
				showPreference();
			} else if( ke.getKeyCode() == KeyEvent.VK_PERIOD ) {
				// alt-. 開啟站台管理
				showSiteManager();
			} else {
				// 雖然按了 alt 或 meta, 但不認識，
				// 繼續往下送。
				return false;
			}

			// 功能鍵不再往下送
			return true;
		}
		
		// 一般鍵，繼續往下送。
		return false;
	}

	public void keyTyped(KeyEvent e) {}

	public void keyPressed(KeyEvent e) {
		if( e.getSource() == siteField ) {
			if( e.getKeyCode() == KeyEvent.VK_ESCAPE ) {
				updateTab();
			}
		}
	}

	public void keyReleased(KeyEvent e) {}
	
	public void componentResized(ComponentEvent ce)
	{
		updateSize();
	}

	public void componentMoved(ComponentEvent ce) {}
	public void componentShown(ComponentEvent ce) {}
	public void componentHidden(ComponentEvent ce) {}

	public ZTerm()
	{
		// FIXME: magic number
		super("ZTerm");
	
		// 先用預設值初始化設定，若是在設定檔中有設定則會取代預設值。
		resource = new Resource();
		resource.initValue();
		resource.readFile();

		conv = new Convertor();
		
		// 各個連線
		sessions = new Vector();
		
		// 與系統剪貼簿溝通的橋樑
		clip = new Clip();
		
		// 設定 Look and Feel
		updateLookAndFeel();
		
		// 初始化各種 icon
		connectedIcon = new ImageIcon( ZTerm.class.getResource( "icon/connected.png" ) );
		closedIcon = new ImageIcon( ZTerm.class.getResource( "icon/closed.png" ) );
		bellIcon = new ImageIcon( ZTerm.class.getResource( "icon/bell.png" ) );
		
		// 是否要顯示工具列
		showToolbar = resource.getBooleanValue( Resource.SHOW_TOOLBAR );
		
		// 設定主畫面 Layout
		getContentPane().setLayout( new BorderLayout() );
		
		makeMenu();
		makePopupMenu();
		makeTabbedPane();
		makeControlPanel();
		
		// 設定視窗位置、大小
		updateBounds();
		// 設定好視窗大小後才知道 image 大小
		bi = new BufferedImage( width, height, BufferedImage.TYPE_INT_RGB );
		
		setVisible( true );
		setResizable( true );
		
		// 一開始沒有任何 dialog
		isShowingDialog = false;
		
		// 按下關閉視窗按鈕時只執行 quit()
		setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
		addWindowListener( new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				quit();
			}
		});
		
		// 接收視窗大小改變的事件
		addComponentListener( this );
		
		// 攔截鍵盤 event 以處理快速鍵
		DefaultKeyboardFocusManager.getCurrentKeyboardFocusManager().addKeyEventDispatcher(this);
		
		// 自動連線
		autoconnect();
	}
	
	public static void main( String args[] ) {
		new ZTerm();
	}

}
